# Generated by Django 3.2.19 on 2023-06-06 09:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0009_preset_base_model"),
    ]

    operations = [
        migrations.AddField(
            model_name="annotationfield",
            name="required",
            field=models.BooleanField(default=False),
        ),
    ]
